/*
 *  tnes.c
 *  v1
 *
 *  NES to TNES converter by hippy dave, utility functions borrowed from makerom by 3DSGuy.
 *
 */

#include "tnes.h"

typedef struct
{
	u8 ident[4];	//= { 0x54, 0x4e, 0x45, 0x53 };
	u8 mapper;		//= 0;
	u8 prg;			//= 0;
	u8 chr;			//= 0;
	u8 sram;		//= 0;
	u8 mirror;		//= 0;
	u8 vram;		//= 0;
	u8 unused[6];	//= { 0, 0, 0, 0, 0, 0 };
} tnes_header;

u8 nes_header[4] = { 0x4e, 0x45, 0x53, 0x1a };

int main(int argc, char *argv[])
{
	fprintf(stdout,"NES to TNES converter by hippy dave, utility functions borrowed from makerom by 3DSGuy.\n");
	if(argc != 3) {
		fprintf(stderr,"Usage: %s <nesinput.nes> <tnesoutput.nes>\n",argv[0]);
		return -1;
	}

	tnes_header thead;
	thead.ident[0] = 0x54;
	thead.ident[1] = 0x4e;
	thead.ident[2] = 0x45;
	thead.ident[3] = 0x53;
	//
	u64 fsize = GetFileSize64(argv[1]);
	u8* data = ImportFile(argv[1], 0);
	bool nesheaderok = true;
	int i;
	for(i=0; i<4; i++){
		if(data[i]!= nes_header[i])
			nesheaderok = false;
	}
	
	if(!nesheaderok){
		fprintf(stderr,"Error: NES header not found in input file, aborting!\n");
		free(data);
		return -2;
	}
	
	u8 mapno = data[6] & 0xF0;
	mapno >>= 4;
	u8 maphi = data[7] & 0xF0;
	mapno |= maphi;
	fprintf(stdout,"mapper number from NES header: %i - ",mapno);
	
	switch(mapno){
		case 0:	//none
			thead.mapper = 0;
			fprintf(stdout,"none\n");
			break;
		case 1:	//mmc1
			thead.mapper = 1;
			fprintf(stdout,"mmc1\n");
			break;
		case 4:	//mmc3
			thead.mapper = 3;
			fprintf(stdout,"mmc3\n");
			break;
		case 5:	//mmc5
			thead.mapper = 5;
			fprintf(stdout,"mmc5\n");
			break;
		case 2: //unrom
			thead.mapper = 6;
			fprintf(stdout,"unrom\n");
			break;
		case 71: //camerica
			thead.mapper = 6;
			fprintf(stdout,"camerica - patching to unrom, may cause issues\n");
			break;
		default: //unsupported
			fprintf(stderr,"not currently supported, aborting! (Error)\n");
			free(data);
			return -3;
			break;
	}
	
	thead.prg = data[4] << 1;
	thead.chr = data[5];
	thead.sram = (data[6] & 2) >> 1;
	if((data[6] & 8) == 8){
		thead.mirror = 0;
	} else {
		if ((data[6] & 1) == 0){
			thead.mirror = 1;
		} else {
			thead.mirror = 2;
		}
	}
	thead.vram = 0;	//dunno tbh.
	
	if ((data[7] & 0xc0) == 0xc0)
		fprintf(stdout,"Warning: NES 2.0 header detected, may not be fully supported.\n");
	
	//must be a neater way to copy the structure in one go
	memcpy(data,thead.ident,sizeof thead.ident);
	data[4] = thead.mapper;
	data[5] = thead.prg;
	data[6] = thead.chr;
	data[7] = thead.sram;
	data[8] = thead.mirror;
	data[9] = thead.vram;
	
	FILE *outfile = fopen(argv[2],"wb");
	WriteBuffer(data, fsize, 0, outfile);
	

	//
	free(data);
	fclose(outfile);
	fprintf(stdout,"Done!?\n");
	
	return 0;
}



// from Project_CTR/makerom/utils.c by 3DSGuy and any other contributors

u64 GetFileSize64(char *filename)
{
#ifdef _WIN32
	struct _stat64 st;
	if( _stat64(filename, &st) != 0)
		return 0;
	else
		return st.st_size;
#else
	struct stat st;
	if( stat(filename, &st) != 0)
		return 0;
	else
		return st.st_size;
#endif
}

u8* ImportFile(char *file, u64 size)
{
	u64 fsize = GetFileSize64(file);
	if(size > 0 && size != fsize){
		fprintf(stderr,"[!] %s has an invalid size (0x%"PRIx64")\n",file, fsize);
		return NULL;
	}
	
	u8 *data = (u8*)calloc(1,fsize);
	if(!data){
		fprintf(stderr,"[!] Not enough memory\n");
		return NULL;
	}
	FILE *fp = fopen(file,"rb");
	fread(data,fsize,1,fp);
	fclose(fp);
	
	return data;
}

void WriteBuffer(void *buffer, u64 size, u64 offset, FILE *output)
{
	fseek_64(output,offset);
	fwrite(buffer,size,1,output);
} 

int fseek_64(FILE *fp, u64 file_pos)
{
#ifdef _WIN32
	fpos_t pos = file_pos;
	return fsetpos(fp,&pos);
#else
	return fseeko(fp,file_pos,SEEK_SET);
#endif
}
